{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import plotly.graph_objs as go\n",
    "import plotly.offline as py\n",
    "import glob\n",
    "import numpy as np\n",
    "py.init_notebook_mode()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "First, a few utilities:\n",
    "* `Data` is a class to contain data to be plotted and ridden from files\n",
    "* `MyData` is a class which implements the reading from files (usually in the constructor)\n",
    "* `open_files` is a function open files with data to be plotted\n",
    "\n",
    "You can find the general schema here: https://vecta.io/app/edit/-Laa83yl4nPcFXhrVpS3"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Data():\n",
    "    \"\"\"\n",
    "    Generic class to contain data\n",
    "    \"\"\"\n",
    "\n",
    "    def __init__(self, lines, name):\n",
    "        \"\"\"\n",
    "        Retrieve data from a list of text lines\n",
    "        \"\"\"\n",
    "        self._name = name\n",
    "\n",
    "    @property\n",
    "    def name(self):\n",
    "        return self._name\n",
    "\n",
    "    @property\n",
    "    def fields(self):\n",
    "        return self._fields\n",
    "\n",
    "    @property\n",
    "    def values(self):\n",
    "        return self._values\n",
    "\n",
    "    @values.setter\n",
    "    def values(self, val):\n",
    "        self._values = val\n",
    "\n",
    "    @property\n",
    "    def all_data(self):\n",
    "        assert len(self.__fields) == len(\n",
    "            self._values), \"Wrong data initialization, fields and values must have the same size\"\n",
    "        return list(zip(self._fields, self._values))\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "####################################\n",
    "# REIMPLEMENT THIS CLASS IF NEEDED #\n",
    "####################################\n",
    "\n",
    "class MyData(Data):\n",
    "    \"\"\"\n",
    "    Class to contain Precision-Recall and F-measure data\n",
    "    \"\"\"\n",
    "\n",
    "    def __init__(self, lines, name):\n",
    "        super().__init__(lines, name)\n",
    "\n",
    "        cleaned = [line.rstrip('\\n') for line in lines]\n",
    "\n",
    "        # looking for the number of pieces\n",
    "        max_index = 0\n",
    "        for line in cleaned:\n",
    "            if line.startswith('Piece number:'):\n",
    "                max_index = max(max_index, int(line[13:]))\n",
    "\n",
    "        self._values = np.zeros((3, max_index + 1))\n",
    "        self._fields = ['Precision', 'Recall', 'F-measure']\n",
    "\n",
    "        # storing data\n",
    "        for line in cleaned:\n",
    "            if line.startswith('Piece number:'):\n",
    "                index = int(line[13:])\n",
    "            elif line.startswith('Precision:'):\n",
    "                self._values[0, index] = float(line[10:])\n",
    "            elif line.startswith('Recall:'):\n",
    "                self._values[1, index] = float(line[7:])\n",
    "            elif line.startswith('F1-measure:'):\n",
    "                self._values[2, index] = float(line[11:])\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def split_fname(fname, leave_dataset_name=False):\n",
    "    \"\"\"\n",
    "    Split a file name removing extensions and adding white spaces\n",
    "    \"\"\"\n",
    "    fname = fname.replace('_', ' ')\n",
    "    fname = fname.replace('results', '')\n",
    "    fname = fname.replace('validation', '')\n",
    "    fname = fname.replace('-', '')\n",
    "    \n",
    "    if leave_dataset_name:\n",
    "        fname = fname.replace('mozart', 'mozart ')\n",
    "        fname = fname.replace('pop', 'pop ')\n",
    "    else: \n",
    "        fname = fname.replace('mozart', '')\n",
    "        fname = fname.replace('pop', '')\n",
    "    return fname[:fname.find('.')]\n",
    "\n",
    "\n",
    "def open_files(file_names, leave_dataset_name=False):\n",
    "    \"\"\"\n",
    "    Open files and retrieve data\n",
    "\n",
    "    PARAMS:\n",
    "        `file_names`: a list of string consisting of file names\n",
    "\n",
    "    RETURNS:\n",
    "        a Data object\n",
    "    \"\"\"\n",
    "\n",
    "    data = []\n",
    "\n",
    "    for fname in file_names:\n",
    "        with open(fname, 'r') as f:\n",
    "            lines = f.readlines()\n",
    "            data_name = split_fname(fname, leave_dataset_name=leave_dataset_name)\n",
    "            data.append(MyData(lines, data_name))\n",
    "\n",
    "    return data\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Plots of CNN crossvalidation on Mozart and Pop dataset\n",
    "* `cl1`: threshold chosen as the upper sample of the 2nd cluster out of 5\n",
    "* `cl2`: threshold chosen as the upper sample of the 2nd cluster out of 6\n",
    "* non-specified: threshold chosen as the upper sample of the 1st cluster out of 2\n",
    "* mono: melody detected by using the graph search\n",
    "\n",
    "The following is the function used to plot.\n",
    "\n",
    "Violin plot show a boxplot with the mean in white, the percentiles the confidence interval at 95% and the samples outside it. Also, they show the distribution of the samples in the width. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from plotly.io import write_image\n",
    "from scipy.stats import wilcoxon, friedmanchisquare\n",
    "def plot(data, static=False, title=''):\n",
    "    \"\"\"\n",
    "    Plot the data.\n",
    "\n",
    "    Params:\n",
    "        `data`: a list of Data objects\n",
    "        `static`: if True save to SVG\n",
    "        `inline`: if True tries to plot in Jupyter Notebook\n",
    "    \"\"\"\n",
    "\n",
    "    if len(data) == 2:\n",
    "        stat_test = wilcoxon\n",
    "    else:\n",
    "        stat_test = friedmanchisquare\n",
    "\n",
    "    _t = []\n",
    "    for i in range(len(data[0].fields)):\n",
    "        _t.append([data[j].values[i] for j in range(len(data))])\n",
    "        pvalue = [f'{stat_test(*arg)[1]:.3E}' for arg in _t]\n",
    "\n",
    "    traces = []\n",
    "    for i, d in enumerate(data):\n",
    "        x = np.ravel([d.fields] * d.values.shape[1])\n",
    "        y = np.ravel(d.values)\n",
    "\n",
    "        traces.append(\n",
    "            dict(\n",
    "                type='violin',\n",
    "                x=x,\n",
    "                y=y,\n",
    "                marker=dict(\n",
    "                    line=dict(\n",
    "                        color='black',\n",
    "                        width=1.5),\n",
    "                ),\n",
    "                opacity=0.6,\n",
    "                name=d.name,\n",
    "                box=dict(\n",
    "                    visible=True,\n",
    "                    line=dict(\n",
    "                        color='black',\n",
    "                        width=1.5\n",
    "                    )\n",
    "                ),\n",
    "                meanline=dict(\n",
    "                    visible=True,\n",
    "                    color='white',\n",
    "                    width=1.5\n",
    "                ),\n",
    "                spanmode='hard',\n",
    "                scalemode='count'\n",
    "            )\n",
    "        )\n",
    "\n",
    "    layout = go.Layout(\n",
    "        xaxis=dict(\n",
    "            tickfont=dict(\n",
    "                size=18,\n",
    "            )\n",
    "        ),\n",
    "        violinmode='group',\n",
    "        yaxis=dict(\n",
    "            zeroline=False\n",
    "        ),\n",
    "        annotations=[\n",
    "            dict(\n",
    "                x=label,\n",
    "                y=np.min([d.values for d in data]),\n",
    "                text='p = ' + pvalue[i],\n",
    "                showarrow=True,\n",
    "                ax=0,\n",
    "                ay=30\n",
    "            ) for i, label in enumerate(data[0].fields)\n",
    "        ],\n",
    "        title=title\n",
    "    )\n",
    "    fig = go.Figure(data=traces, layout=layout)\n",
    "\n",
    "    if static:\n",
    "        write_image(fig, 'output.svg')\n",
    "    else:\n",
    "        py.iplot(fig)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "# Change this to save SVG files...\n",
    "SAVE = False\n",
    "\n",
    "file_names = sorted(glob.glob('*results*mozart*'))\n",
    "# file_names = file_names[1:]\n",
    "data = open_files(file_names)\n",
    "#plot(data, static=SAVE, title='Mozart crossvalidation')\n",
    "\n",
    "file_names = sorted(glob.glob('*results*pop*'))\n",
    "# file_names = file_names[1:]\n",
    "data = open_files(file_names)\n",
    "plot(data, static=SAVE, title='Pop crossvalidation')"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Validation\n",
    "We validated the models trained on the Pop and Mozart datasets on various composers. The number in parenthesis is the number of works available for each composer."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "SAVE=True\n",
    "file_names = sorted(glob.glob('*validation*'))\n",
    "data = open_files(file_names, leave_dataset_name=True)\n",
    "plot(data, static=SAVE, title='Validation')\n",
    "\n",
    "# Computing averages\n",
    "groups = []\n",
    "xticks = []\n",
    "for i in sorted(glob.glob('../data/solo-accompaniment-dataset/validation/*')):\n",
    "    j = len(glob.glob(i + '/*.pyc.bz'))\n",
    "    if j > 0:\n",
    "        groups.append(j)\n",
    "        xticks.append(i[49:] + ' (' + str(j) + ')')\n",
    "\n",
    "for d in data:\n",
    "    new_dvalues = np.zeros((d.values.shape[0], len(groups)))\n",
    "    k = 0\n",
    "    for i, j in enumerate(groups):\n",
    "        new_dvalues[:, i] = np.mean(d.values[:, k:k+j], axis=1)\n",
    "        k += j\n",
    "    d.values = new_dvalues\n",
    "    \n",
    "traces = []\n",
    "colors = ['blue', 'orange', 'green', 'red', 'black', 'violet']\n",
    "axis = [('x1', 'y1'), ('x2', 'y2'), ('x3', 'y3')]\n",
    "for m in range(data[0].values.shape[0]):\n",
    "    for i, d in enumerate(data):\n",
    "        x = list(range(d.values.shape[1]))\n",
    "        y = d.values[m, :]\n",
    "\n",
    "        traces.append(\n",
    "            dict(\n",
    "                type='scatter',\n",
    "                mode='lines+markers',\n",
    "                x=xticks,\n",
    "                y=y,\n",
    "                xaxis=axis[m][0],\n",
    "                yaxis=axis[m][1],\n",
    "                opacity=0.4,\n",
    "                name=d.name,\n",
    "                marker=dict(\n",
    "                    color=colors[i],\n",
    "                ),\n",
    "                showlegend=not m\n",
    "            )\n",
    "        )\n",
    "\n",
    "layout = dict(\n",
    "    xaxis=dict(\n",
    "        domain=[0, 0.49],\n",
    "        anchor=axis[0][1], \n",
    "        tickangle=45,\n",
    "        tickfont=dict(\n",
    "            size=7\n",
    "        ),\n",
    "        side='top'\n",
    "    ),\n",
    "    xaxis2=dict(\n",
    "        domain=[0, 0.49],\n",
    "        anchor=axis[1][1],\n",
    "        tickangle=-45,\n",
    "        tickfont=dict(\n",
    "            size=7\n",
    "        ),\n",
    "    ),\n",
    "    xaxis3=dict(\n",
    "        domain=[0.51, 1],\n",
    "        anchor=axis[2][1],\n",
    "        title='Average F-measure',\n",
    "        tickangle=-45,\n",
    "        tickfont=dict(\n",
    "            size=7\n",
    "        ),\n",
    "    ),\n",
    "    yaxis=dict(\n",
    "        domain=[0.51, 1],\n",
    "        anchor=axis[0][0],\n",
    "        title='Average Precision',\n",
    "        zeroline=False,\n",
    "        dtick=0.1,\n",
    "    ),\n",
    "    yaxis2=dict(\n",
    "        domain=[0, 0.49],\n",
    "        anchor=axis[1][0],\n",
    "        title='Average Recall',\n",
    "        zeroline=False,\n",
    "        dtick=0.1,\n",
    "    ),\n",
    "    yaxis3=dict(\n",
    "        domain=[0, 1],\n",
    "        anchor=axis[2][0],\n",
    "        zeroline=False,\n",
    "        dtick=0.1,\n",
    "    ),\n",
    ")\n",
    "fig = go.Figure(data=traces, layout=layout)\n",
    "\n",
    "if SAVE:\n",
    "    write_image(fig, 'output.svg')\n",
    "py.iplot(fig)\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Output distributions\n",
    "\n",
    "These are the input and output pianorolls created with the mozart model. Input uses yellow for the ground-truth melody and grey for the accompaniment. Output uses a heatmap."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "from plotly import tools\n",
    "BINSIZE = 0.05\n",
    "\n",
    "def iqr(ys):\n",
    "    quartile_1, quartile_3 = np.percentile(ys, [25, 75])\n",
    "    iqr = quartile_3 - quartile_1\n",
    "    lower_bound = quartile_1 - (iqr * 1.5)\n",
    "    upper_bound = quartile_3 + (iqr * 1.5)\n",
    "    return np.where((ys < upper_bound) | (ys > lower_bound))[0]\n",
    "\n",
    "\n",
    "def crop(dat, x=True, y=True, ref=None):\n",
    "    if ref is None:\n",
    "        ref = dat\n",
    "    # argwhere will give you the coordinates of every non-zero point\n",
    "    true_points = np.argwhere(ref)\n",
    "    # take the smallest points and use them as the top left of your crop\n",
    "    top_left = true_points.min(axis=0)\n",
    "    # take the largest points and use them as the bottom right of your crop\n",
    "    bottom_right = true_points.max(axis=0)\n",
    "    if x:\n",
    "        dat = dat[:, top_left[1]:bottom_right[1]+1]\n",
    "    if y:\n",
    "        dat = dat[top_left[0]:bottom_right[0]+1,:]\n",
    "    \n",
    "    return dat\n",
    "\n",
    "\n",
    "MIN_TH = 1e-15\n",
    "THRESHOLD = 0.5\n",
    "def set_threshold(arr, CLUSTERING='centroid'):\n",
    "    print(\"starting clustering\")\n",
    "    arr = arr.reshape(-1)\n",
    "    arr = arr[arr > MIN_TH]\n",
    "    N_CLUSTER = 2\n",
    "    target_cluster = 1\n",
    "\n",
    "    arr = arr[iqr(arr)]\n",
    "\n",
    "    if CLUSTERING == 'kmeans':\n",
    "        from sklearn.cluster import KMeans\n",
    "        kmeans = KMeans(n_clusters=N_CLUSTER,\n",
    "                        init=np.array([MIN_TH, arr.max()]).reshape(-1, 1))\n",
    "\n",
    "        labels = kmeans.fit_predict(arr.reshape(-1, 1))\n",
    "    else:\n",
    "        import fastcluster\n",
    "        from scipy.cluster.hierarchy import fcluster\n",
    "        from scipy.spatial.distance import pdist\n",
    "\n",
    "        Z = pdist(arr.reshape(-1, 1))\n",
    "        if CLUSTERING == 'single':\n",
    "            X = fastcluster.single(Z)\n",
    "        elif CLUSTERING == 'average':\n",
    "            X = fastcluster.average(Z)\n",
    "        elif CLUSTERING == 'centroid':\n",
    "            X = fastcluster.centroid(Z)\n",
    "        else:\n",
    "            return THRESHOLD\n",
    "\n",
    "        labels = N_CLUSTER - fcluster(X, N_CLUSTER, 'maxclust')\n",
    "\n",
    "    # setting 0 for the minimum cluster\n",
    "    # np.ma.masked_array returns only values where the mask is 0\n",
    "    index = {}\n",
    "    for i, l in enumerate(labels):\n",
    "        index[l] = arr[i]\n",
    "        if len(index.keys()) == N_CLUSTER:\n",
    "            break\n",
    "\n",
    "    index = sorted(index.items(), key=lambda kv: kv[1]) # list of tuples sorted by values\n",
    "    target_label = index[target_cluster - 1][0] # the label of the desired cluster\n",
    "    th = np.max(arr[np.flatnonzero(labels == target_label)]) # max of the down cluster\n",
    "    print(\"found threshold: \" + str(th))\n",
    "    # print(str(np.ma.masked_array(arr, 1 - labels).min()))\n",
    "\n",
    "    return th\n",
    "\n",
    "\n",
    "def plot_pianoroll(fname, x0=0, x1=-1, color='YlGnBu'):\n",
    "    with np.load(fname) as data:\n",
    "        in_pianoroll = data['in_pianoroll']\n",
    "        out_pianoroll = data['out_pianoroll']\n",
    "        melody = data['melody']\n",
    "        saliency = data['saliency'] \n",
    "    \n",
    "    saliency *= out_pianoroll\n",
    "    # saliency += np.abs(saliency.min())\n",
    "    # saliency *= out_pianoroll\n",
    "    # saliency *= 1 / saliency.max()\n",
    "    data = [\n",
    "        dict(\n",
    "            type='heatmap',\n",
    "            z=crop((in_pianoroll + melody)/2)[:, x0:x1],\n",
    "            colorscale='Hot',\n",
    "            showscale=False,\n",
    "            y=[np.argwhere(in_pianoroll).min(axis=0)[0]]\n",
    "        ),\n",
    "        dict(\n",
    "            type='heatmap',\n",
    "            z=crop(out_pianoroll)[:, x0:x1],\n",
    "            colorscale='Reds',\n",
    "            showscale=True,\n",
    "            colorbar=dict(\n",
    "                len=0.66,\n",
    "                y=0.66,\n",
    "            ),\n",
    "            y=[np.argwhere(out_pianoroll).min(axis=0)[0]],\n",
    "        ),\n",
    "        dict(\n",
    "            type='heatmap',\n",
    "            z=crop(saliency,ref=out_pianoroll)[:, x0:x1],\n",
    "            colorscale=color,\n",
    "            showscale=True,\n",
    "            colorbar=dict(\n",
    "                len=0.33,\n",
    "                y=0.165,\n",
    "            ),\n",
    "            y=[np.argwhere(out_pianoroll).min(axis=0)[0]]\n",
    "        ),\n",
    "    ]\n",
    "\n",
    "    fig = tools.make_subplots(\n",
    "        rows=3, \n",
    "        cols=1,\n",
    "        shared_yaxes=True, \n",
    "        shared_xaxes=True, \n",
    "        vertical_spacing=0.01, \n",
    "        horizontal_spacing=0.01,\n",
    "    )\n",
    "    fig.append_trace(data[0], 1, 1)\n",
    "    fig.append_trace(data[1], 2, 1)\n",
    "    fig.append_trace(data[2], 3, 1)\n",
    "    if SAVE:\n",
    "        write_image(fig, 'output.svg')\n",
    "    print(\"Pianorolls of \" + fname)\n",
    "    py.iplot(fig)\n",
    "        \n",
    "        \n",
    "def plot_histograms(fname, x0=0, x1=-1):\n",
    "    with np.load(fname) as data:\n",
    "        in_pianoroll = data['in_pianoroll']\n",
    "        out_pianoroll = crop(data['out_pianoroll'], ref=in_pianoroll)\n",
    "        melody = crop(data['melody'], ref=in_pianoroll) * out_pianoroll\n",
    "        \n",
    "    out_pianoroll = out_pianoroll[:, x0:x1]\n",
    "    melody = melody[:, x0:x1]\n",
    "    threshold = set_threshold(out_pianoroll)\n",
    "        \n",
    "    data = [\n",
    "        dict(\n",
    "            type='histogram',\n",
    "            x=out_pianoroll[out_pianoroll > 1e-15],\n",
    "            # histnorm='probability',\n",
    "            # opacity=0.6,\n",
    "            xbins=dict(\n",
    "                size=BINSIZE,\n",
    "            ),\n",
    "            name='Any pixel > 0'\n",
    "        ),\n",
    "        dict(\n",
    "            type='histogram',\n",
    "            x=melody[melody > 1e-15],\n",
    "            #histnorm='probability',\n",
    "            #opacity=0.6,\n",
    "            xbins=dict(\n",
    "                size=BINSIZE,\n",
    "            ),\n",
    "            name='Melody pixels'\n",
    "        ),\n",
    "    ]\n",
    "\n",
    "    hist, _ = np.histogram(data[0]['x'], bins=np.arange(-BINSIZE/2, 1+BINSIZE, BINSIZE))\n",
    "    max_hist_value = np.max(hist)\n",
    "\n",
    "    layout = dict(\n",
    "        barmode='overlay',\n",
    "        title='Distribution of predicted values for certain pixels',\n",
    "        shapes=[\n",
    "            # Line Vertical\n",
    "            dict(\n",
    "                type='line',\n",
    "                x0=threshold,\n",
    "                y0=0,\n",
    "                x1=threshold,\n",
    "                y1=max_hist_value,\n",
    "                line=dict(\n",
    "                    color='red',\n",
    "                    width=3\n",
    "                ),\n",
    "            ),\n",
    "        ],\n",
    "    )\n",
    "\n",
    "    fig = go.Figure(data=data, layout=layout)\n",
    "    if SAVE:\n",
    "        write_image(fig, 'output.svg')\n",
    "    print(\"Histograms of \" + fname)\n",
    "    py.iplot(fig)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "SAVE=True\n",
    "plot_pianoroll('../inspection/gluck_die_sommernacht.npz')\n",
    "plot_histograms('../inspection/gluck_die_sommernacht.npz')\n",
    "\n",
    "plot_pianoroll('../inspection/albeniz_tango.npz')\n",
    "plot_histograms('../inspection/albeniz_tango.npz')\n",
    "\n",
    "plot_pianoroll('../inspection/liszt_die_glocken_von_marling.npz')\n",
    "plot_histograms('../inspection/liszt_die_glocken_von_marling.npz')\n",
    "\n",
    "plot_pianoroll('../inspection/schubert_ave_maria.npz', x0=380, x1=620)\n",
    "plot_histograms('../inspection/schubert_ave_maria.npz', x0=380, x1=620)\n",
    "\n",
    "plot_pianoroll('../inspection/mozart_kv475_1.npz', x0=200, x1=700)\n",
    "plot_histograms('../inspection/mozart_kv475_1.npz', x0=200, x1=700)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Kernels\n",
    "Let us plot the kernels after the first  and second layer..."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "import pickle\n",
    "\n",
    "def plot_kernels(fname):\n",
    "    kernels = pickle.load(open(fname, 'rb'), encoding='latin1')\n",
    "    fig = tools.make_subplots(\n",
    "        rows=3,\n",
    "        cols=7, \n",
    "        print_grid=False, \n",
    "        shared_yaxes=True, \n",
    "        shared_xaxes=True, \n",
    "        vertical_spacing=0.005, \n",
    "        horizontal_spacing=0.005\n",
    "    )\n",
    "\n",
    "    for i, kernel in enumerate(kernels[0]):\n",
    "        data = dict(\n",
    "                type='heatmap',\n",
    "                z=kernel[0],\n",
    "                colorscale='YlGnBu',\n",
    "                showscale=(i==0)\n",
    "            )\n",
    "\n",
    "        col = int(i%7) + 1\n",
    "        row = int(i/7) + 1\n",
    "        fig.append_trace(data, row, col)\n",
    "\n",
    "    py.iplot(fig)\n",
    "    \n",
    "plot_kernels('../inspection/nn_kernels_mozart.pkl')\n",
    "plot_kernels('../inspection/nn_kernels_pop.pkl')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "def plot_subkernels(fname):\n",
    "    kernels = pickle.load(open(fname, 'rb'), encoding='latin1')\n",
    "    fig = tools.make_subplots(\n",
    "        rows=21,\n",
    "        cols=21, \n",
    "        print_grid=False, \n",
    "        shared_yaxes=True, \n",
    "        shared_xaxes=True, \n",
    "        vertical_spacing=0.001, \n",
    "        horizontal_spacing=0.001\n",
    "    )\n",
    "\n",
    "    for i, kernel in enumerate(kernels[1]):\n",
    "        for j, sub_kernel in enumerate(kernel):\n",
    "            k = i + j\n",
    "            data = dict(\n",
    "                    type='heatmap',\n",
    "                    z=sub_kernel,\n",
    "                    colorscale='YlGnBu',\n",
    "                    showscale=(k==0)\n",
    "                )\n",
    "\n",
    "            col = i + 1\n",
    "            row = j + 1\n",
    "            fig.append_trace(data, row, col)\n",
    "\n",
    "    py.iplot(fig)\n",
    "    \n",
    "plot_subkernels('../inspection/nn_kernels_mozart.pkl')\n",
    "plot_subkernels('../inspection/nn_kernels_pop.pkl')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "plot_pianoroll('./inspection/gluck_masked_saliency.npz', color='Reds')\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "def plot_masked_saliency(fname, note, x0=0, x1=-1):    \n",
    "    with np.load(fname) as data:\n",
    "        in_pianoroll = data['in_pianoroll']\n",
    "        out_pianoroll = data['out_pianoroll']\n",
    "        melody = data['melody']\n",
    "        masked_history = data['masked']\n",
    "    \n",
    "    correction = np.zeros_like(in_pianoroll)\n",
    "    output = np.zeros_like(in_pianoroll)\n",
    "    for iteration in masked_history:\n",
    "        prediction = iteration[0]\n",
    "        mask = iteration[1]\n",
    "        if np.all(mask[note[0], note[1]:note[2]]):\n",
    "            correction[mask == 0] += 1\n",
    "            output[mask == 0] += np.mean(\n",
    "                out_pianoroll[note[0], note[1]:note[2]] - prediction[note[0], note[1]:note[2]]\n",
    "            )\n",
    "    \n",
    "    output /= np.clip(correction, 1e-15, None, out=correction)\n",
    "    output /= np.abs(output).max()\n",
    "    output *= in_pianoroll\n",
    "        \n",
    "    data = [\n",
    "        dict(\n",
    "            type='heatmap',\n",
    "            z=crop((in_pianoroll + melody)/2)[:, x0:x1],\n",
    "            colorscale='Hot',\n",
    "            showscale=False,\n",
    "            y=[np.argwhere(in_pianoroll).min(axis=0)[0]]\n",
    "        ),\n",
    "        dict(\n",
    "            type='heatmap',\n",
    "            z=crop(out_pianoroll)[:, x0:x1],\n",
    "            colorscale='Reds',\n",
    "            showscale=True,\n",
    "            colorbar=dict(\n",
    "                len=0.33,\n",
    "                y=0.85,\n",
    "            ),\n",
    "            y=[np.argwhere(out_pianoroll).min(axis=0)[0]],\n",
    "        ),\n",
    "        dict(\n",
    "            type='heatmap',\n",
    "            z=crop(output,ref=out_pianoroll)[:, x0:x1],\n",
    "            colorscale='YlGnBu',\n",
    "            showscale=True,\n",
    "            colorbar=dict(\n",
    "                len=0.66,\n",
    "                y=0.33,\n",
    "            ),\n",
    "            y=[np.argwhere(out_pianoroll).min(axis=0)[0]]\n",
    "        ),\n",
    "    ]\n",
    "\n",
    "    fig = tools.make_subplots(\n",
    "        rows=3, \n",
    "        cols=1,\n",
    "        shared_yaxes=True, \n",
    "        shared_xaxes=True, \n",
    "        vertical_spacing=0.01, \n",
    "        horizontal_spacing=0.01,\n",
    "    )\n",
    "    \n",
    "    fig.append_trace(data[0], 1, 1)\n",
    "    fig.append_trace(data[1], 2, 1)\n",
    "    fig.append_trace(data[2], 3, 1)\n",
    "    fig.layout.shapes=[\n",
    "    # unfilled Rectangle\n",
    "        dict(\n",
    "            type='rect',\n",
    "            xref='x1',\n",
    "            yref='y1',\n",
    "            x0=note[1]-0.49,\n",
    "            y0=note[0]+1,\n",
    "            x1=note[2]-0.49,\n",
    "            y1=note[0]-1,\n",
    "            line=dict(\n",
    "                color='green',\n",
    "            ),\n",
    "        ),\n",
    "        dict(\n",
    "            type='rect',\n",
    "            xref='x1',\n",
    "            yref='y2',\n",
    "            x0=note[1]-0.49,\n",
    "            y0=note[0]+1,\n",
    "            x1=note[2]-0.49,\n",
    "            y1=note[0]-1,\n",
    "            line=dict(\n",
    "                color='green',\n",
    "            ),\n",
    "        ),\n",
    "        dict(\n",
    "            type='rect',\n",
    "            xref='x1',\n",
    "            yref='y3',\n",
    "            x0=note[1]-0.49,\n",
    "            y0=note[0]+1,\n",
    "            x1=note[2]-0.49,\n",
    "            y1=note[0]-1,\n",
    "            line=dict(\n",
    "                color='green',\n",
    "            ),\n",
    "        ),\n",
    "    ]\n",
    "\n",
    "    if SAVE:\n",
    "        write_image(fig, 'output.svg')\n",
    "    print(\"Pianorolls of \" + fname)\n",
    "    py.iplot(fig)\n",
    "\n",
    "plot_masked_saliency('../inspection/gluck_masked_saliency.npz', (59, 60, 63))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import sys\n",
    "import os\n",
    "sys.path.append(os.path.abspath('../../convnet_drawer/'))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from convnet_drawer import Model, Conv2D, Deconv2D\n",
    "\n",
    "model = Model(input_shape=(128, 64, 1))\n",
    "model.add(Conv2D(21, (32, 16), (1, 1), padding=\"valid\"))\n",
    "# model.add(MaxPooling2D((3, 3), strides=(2, 2)))\n",
    "model.add(Conv2D(441, (32, 16), (1,1), padding=\"valid\"))\n",
    "# model.add(MaxPooling2D((3, 3), strides=(2, 2)))\n",
    "model.add(Deconv2D(21, (32, 16), (1, 1), padding=\"valid\"))\n",
    "model.add(Deconv2D(1, (32, 16), padding=\"valid\"))\n",
    "\n",
    "\n",
    "# save as svg file\n",
    "model.save_fig(\"network.svg\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Network structure\n",
    "\n",
    "To see the network structure after editing by hand go to https://vecta.io/app/edit/-La_LUq_4Fld0mlpt2lq\n",
    "\n",
    "## Graph building\n",
    "Image: https://vecta.io/app/edit/-La_auXXws0bpTiDvVUf\n",
    "The graph is built in this way:\n",
    "1. Create the starting and ending virtual notes (green) and put them at the top and at the bottom of the list of notes\n",
    "2. Select note 0\n",
    "3. Set `last offset` = 0\n",
    "4. Create a branch from the selected note to all the notes with the minor onset so that:\n",
    "    1. the onset is also $\\ge$ than `last offset`\n",
    "    2. the probability of the note is $\\gt$ threshold (red notes)\n",
    "5. The weight of the branch is $-p$, where $p$ is the probability of the arriving note\n",
    "6. If there are no notes which satisfy condition `5.`, connect this note to the virtual last note\n",
    "7. Select the next note with probability $\\gt$ than threshold and restart from point `4.`\n",
    "\n",
    "In the image, green notes are virtual nodes, red notes are notes over threshold, yellow notes are notes under threshold and azure notes are notes over threshold that can't be reached by any path.\n",
    "\n",
    "## General schema\n",
    "You can find the general schema here: https://vecta.io/app/edit/-Laa83yl4nPcFXhrVpS3"
   ]
  }
 ],
 "metadata": {
  "celltoolbar": "Tags",
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
